#include "mbed.h"
#include "./Display/display.h"      // LCD Display Driver
#include "./led/led.h"              // LED Driver
#include "./Input/input.h"          // INPUT Driver
#include "./Sound/sound.h"          // SOUND Driver

#pragma once

enum GAME_MODE
{
    eGameTitle,
    eGamePlay,
    eGameOver,
};

#define FRAMES_PER_SECOND   20      // unfortunately this is what we get...
#define MS_PER_FRAME        50

class Game
{   
    int frame_timeout;
    int frame_count;
    int now;

    Timer timer;
    display disp;
    AnalogIn ain;
    led lights;
    input inputs;
    Sound sound;
    
    GAME_MODE mode;
    int game_step;
    bool paused;
    
    void initialize();
    bool dotick();
    
    void DoGameTitle();
    void DoGamePlay();
    void DoGameOver();
    
    uint8_t star_field_timer;
    void SetupStarfield();
    void UpdateStarfield();
    
    void SetupInvaders();
    void UpdateInvaders();
    bool invaders[40];
    int8_t active_invaders;
    int8_t invaders_inc;
    uint8_t invaders_timer;
    uint8_t invaders_wait_time;
    uint8_t invaders_frame;
    uint8_t invaders_x, invaders_y;
    bool invader_active_shot[4];
    uint8_t invader_shot_interval;
    uint8_t invader_shot_x[4];
    uint8_t invader_shot_y[4];
    int8_t invader_shot_timer;
    uint8_t invader_march_note;
    
    
    void SetupPlayer();
    void UpdatePlayer();
    uint8_t player_x, player_y;
    bool player_shot_active;
    uint8_t player_shot_x, player_shot_y;
    
    void SetupSaucer();
    void UpdateSaucer();
    bool saucer_active;
    int saucer_timer;
    int8_t saucer_inc;
    uint8_t saucer_x, saucer_y;
    
    void SetupShields();
    void UpdateShields();
    uint8_t shield_x, shield_y;
    uint8_t shield_points[4];
    
    void CheckCollisions();
    
    void IncScore(int points);
    void UpdateScore();
    void UpdateLives();
    int hiscore;
    int score;
    uint8_t lives;
    uint8_t level;
    
    uint8_t wait_timer;
    
    void UpdateGameOverInvaders();
    uint8_t game_over_msg;
    
public:

    Game();
    
    void tick();
};